''' Properties for modeling Chart inputs, constraints, and dependencies.

selection spec:
    [['x'], ['x', 'y']]
    [{'x': categorical, 'y': numerical}]

'''
from __future__ import absolute_import

import numpy as np
import pandas as pd

from bokeh.core.has_props import HasProps
from bokeh.core.property.bases import PrimitiveProperty
from bokeh.core.properties import Array, bokeh_integer_types, Bool, Either, Int, List, String

from .utils import special_columns, title_from_columns

class Column(Array):
    """Represents column-oriented data.

    This property is used to provide a consistent interface for column-like data. The
    property both validates the data types set, and transforms all column-like data into
    `pd.Series` data.
    """
    def _is_seq(self, value):
        is_array = super(Column, self)._is_seq(value)
        return (isinstance(value, pd.Series) or isinstance(value, pd.Index) or
                isinstance(value, list) or is_array)

    def _new_instance(self, value):
        return pd.Series(value)

    def transform(self, value):
        if value is None:
            return None

        if isinstance(value, pd.Series):
            arr = value.values
        else:
            arr = pd.Series(value).values

        trans_array = super(Column, self).transform(arr)
        try:
            return pd.Series(trans_array)
        except ValueError:

            raise ValueError("Could not transform %r" % value)


class Logical(Bool):
    """A boolean like data type.

    This property is valid for both python and numpy boolean types.
    """
    def validate(self, value):
        try:
            super(Logical, self).validate(value)
        except ValueError:
            if isinstance(value, list):
                value = np.array(value)

            # If not a Bool, then look for pseudo-logical types
            if isinstance(value, np.ndarray):
                values = np.unique(value)
                if len(values) == 2:
                    return

                raise ValueError('expected a Bool or array with 2 unique values, got %s' % value)


class ColumnLabel(Either):
    """Specify a column by name or index."""

    def __init__(self, columns=None, default=None, help=None):
        # ToDo: make sure we can select by integer
        types = (String, Int)
        self.columns = columns
        super(ColumnLabel, self).__init__(*types, default=default, help=help)

    def validate(self, value):
        """If we are given a column list, make sure that the column provided is valid."""
        super(ColumnLabel, self).validate(value)

        if self.columns:
            if type(value) in bokeh_integer_types:
                if len(self.columns) > value:
                    return
                else:
                    raise ValueError("Not a valid column selection.")
            else:
                if value not in self.columns and value not in special_columns:
                    raise ValueError("Column provided is not in the list of valid columns: %s" % self.columns)

    def __str__(self):
        return "Column Name or Column String"


class Dimension(HasProps):
    """Configures valid Chart column selections.

    A dimension is Chart property that is assigned one or more columns names or indices.
    Each column can match one or more column types, which are important to charts,
    because the type of column selection can greatly affect the behavior of generalized
    Charts.

    The Dimension also provides convenient utilities for accessing information
    about the current provided configuration at the global, non-grouped level.

    The dimension configuration does not require the data, but when the data is
    added using the `set_data` method, then validation can occur of the settings
    by using the `valid` and `invalid` types identified by the selection.
    """

    name = String()
    alt_names = Either(String, List(String), default=None)
    columns = Either(ColumnLabel, List(ColumnLabel), default=None)

    valid = Either(PrimitiveProperty, List(PrimitiveProperty), default=None)
    invalid = Either(PrimitiveProperty, List(PrimitiveProperty), default=None)

    selection = Either(ColumnLabel, List(ColumnLabel), default=None)

    def __init__(self, name, **properties):
        properties['name'] = name
        super(Dimension, self).__init__(**properties)
        self._data = pd.DataFrame()
        self._chart_source = None

    def get_valid_types(self, col_data):
        """Returns all property types that are matched."""
        valid_types = list(self.valid)
        matches = []

        # validate each type on the provided column
        for valid_type in valid_types:
            prop = valid_type()

            # if valid, append to the output
            try:
                prop.validate(col_data)
                matches.append(valid_type)
            except ValueError:
                pass

        return matches

    @property
    def data(self):
        """The data selected for the Dimension.

        Returns pd.Series(1) if data is empty or no selection.
        """
        if self._data.empty or self.selection is None:
            return pd.Series(1)
        else:
            # return special column type if available
            if self.selection in list(special_columns.keys()):
                return special_columns[self.selection](self._data)

            return self._data[self.selection]

    def __len__(self):
        return len(self.data.index)

    def set_data(self, data):
        """Set data property so that builders has access to configuration metadata.

        Args:
            data (`ChartDataSource`): the data source associated with the chart
        """
        self.selection = data[self.name]
        self._chart_source = data
        self._data = data.df
        self.columns = list(self._data.columns.values)

    @property
    def min(self):
        """The minimum of one to many column selections."""
        if isinstance(self.data, pd.Series):
            return self.data.min()
        else:
            return self.data.min(axis=1).min()

    @property
    def max(self):
        """The maximum of one to many column selections."""
        if isinstance(self.data, pd.Series):
            return self.data.max()
        else:
            return self.data.max(axis=1).max()

    @property
    def dtype(self):
        if isinstance(self.data, pd.DataFrame):
            return self.data.dtypes[self.selection[0]]
        else:
            return self.data.dtype

    @property
    def computed(self):
        """Check the `ChartDataSource` to see if the selection is a derived column."""
        if self._chart_source is None:
            return False
        else:
            return self._chart_source.is_computed(self.selection)

    @property
    def selected_title(self):
        """A title formatted representation of selected columns."""
        return title_from_columns(self.selection)


class EitherColumn(Either):
    """Allow providing option of column types."""

    # ToDo: incorporate fix into Either
    def matches(self, new, old):
        comparison = super(EitherColumn, self).matches(new, old)
        if isinstance(comparison, bool):
            return comparison
        elif isinstance(comparison, pd.Series):
            return comparison.all()
        else:
            raise ValueError('Failed when comparing Columns')



"""This is the Bokeh charts testing interface.

"""
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2014, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from __future__ import absolute_import

from mock import Mock
import pytest
from unittest import skipIf

try:
    import pandas as pd
    is_pandas = True
except ImportError as e:
    is_pandas = False

from bkcharts.builder import Builder, XYBuilder
from bkcharts.properties import Dimension
from bkcharts.attributes import ColorAttr, DEFAULT_PALETTE
from bokeh.models import Range1d, FactorRange

#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------


CUSTOM_PALETTE = ['Red', 'Green', 'Blue']


@pytest.fixture
def test_builder():
    class TestBuilder(Builder):
        default_attributes = {'color': ColorAttr()}
        x = Dimension('x')
        y = Dimension('y')

        dimensions = ['x', 'y']

    return TestBuilder


@pytest.fixture
def simple_builder(test_builder, test_data):
    return test_builder(test_data.pd_data)


@pytest.fixture
def custom_palette_builder(test_builder, test_data):
    return test_builder(test_data.pd_data, palette=CUSTOM_PALETTE)


def test_empty_builder_labels(test_builder):
    builder = test_builder()
    assert builder.xlabel is None
    assert builder.ylabel is None


def test_default_color(simple_builder):
    assert simple_builder.attributes['color'].iterable == DEFAULT_PALETTE


def test_custom_color(custom_palette_builder):
    assert custom_palette_builder.attributes['color'].iterable == CUSTOM_PALETTE

def test_legend_sort(test_data):
    col1, col2, col3 = Mock(), Mock(), Mock()
    legends = [('col1', col1), ('col3', col3), ('col2', col2)]
    items = [('col1', Mock()), ('col2', Mock()), ('col3', Mock())]
    attributes = {
        'color': Mock(columns=['series'], items=items),
        'dash': Mock(columns=['series'], items=items),
        'marker': Mock(columns=['series'], items=items)
    }

    # assert we don't change anything if sort_legend is not specified
    assert Builder._sort_legend(None, None, legends, attributes) == legends
    # assert we sort asc if specified so
    assert Builder._sort_legend('color', 'ascending', legends, attributes) == \
        [('col1', col1), ('col2', col2), ('col3', col3)]
    # assert we sort desc if specified so
    assert Builder._sort_legend('color', 'descending', legends, attributes) == \
        [('col3', col3), ('col2', col2), ('col1', col1)]

@skipIf(not is_pandas, "pandas not installed")
def test_created_range_type(test_builder, test_data):
    pd_data = test_data.pd_data.copy()
    pd_data['col3'] = ['a','b','c','b']
    pd_data['col4'] = pd.date_range('1950-01', '1950-05', freq='M')

    builder1 = XYBuilder(pd_data, x='col1')
    assert isinstance(builder1._get_range('x', 1, 4), Range1d)
    assert builder1.xscale == 'linear'

    # "None" as str is for label repr, handles case where there is no x dim
    builder2 = XYBuilder(pd_data, x='col2')
    assert isinstance(builder2._get_range('x', "None", "None"), FactorRange)
    assert builder2.xscale == 'categorical'

    builder3 = XYBuilder(pd_data, x='col3')
    assert isinstance(builder3._get_range('x', 'a', 'c'), FactorRange)
    assert builder3.xscale == 'categorical'

    builder4 = XYBuilder(pd_data, x='col4')
    assert isinstance(builder4._get_range('x', 0, 1), Range1d)
    assert builder4.xscale == 'datetime'

def test_sort_legend(test_builder, test_data):
    test_builder = test_builder(test_data.pd_data, legend_sort_field='color', legend_sort_direction='ascending')

    assert test_builder.legend_sort_field == 'color'
    assert test_builder.legend_sort_direction == 'ascending'



"""This is the Bokeh charts interface. It gives you a high level API
to build complex plot is a simple way.

This is the Area builder which lets you build your Area charts
just passing the arguments to the Chart class and calling the proper
functions.
"""
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2014, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from __future__ import absolute_import

from bkcharts.builder import create_and_build
from bkcharts.glyphs import AreaGlyph
from .line_builder import LineBuilder

#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------


def Area(data=None, x=None, y=None, **kws):
    """ Create an area chart using :class:`AreaBuilder
    <bkcharts.builders.area_builder.AreaBuilder>` to render the
    geometry from values.

    Args:
        data (:ref:`userguide_charts_data_types`): table-like data
        x (str or list(str), optional): the column label to use for the x dimension
        y (str or list(str), optional): the column label to use for the y dimension

    In addition the the parameters specific to this chart,
    :ref:`userguide_charts_defaults` are also accepted as keyword parameters.

    Returns:
        :class:`Chart`: includes glyph renderers that generate the area glyphs

    Examples:

    .. bokeh-plot::
        :source-position: above

        from bkcharts import Area, show, output_file

        # create some example data
        data = dict(
            python=[2, 3, 7, 5, 26, 221, 44, 233, 254, 265, 266, 267, 120, 111],
            pypy=[12, 33, 47, 15, 126, 121, 144, 233, 254, 225, 226, 267, 110, 130],
            jython=[22, 43, 10, 25, 26, 101, 114, 203, 194, 215, 201, 227, 139, 160],
        )

        area = Area(data, title="Area Chart", legend="top_left",
                    xlabel='time', ylabel='memory')

        output_file('area.html')
        show(area)

    """
    kws['x'] = x
    kws['y'] = y
    return create_and_build(AreaBuilder, data, **kws)


class AreaBuilder(LineBuilder):
    """This is the Area builder and it is in charge of generating
    glyph renderers that together produce an area chart.

    Essentially, we provide a way to ingest the data, make the proper
    calculations and push the references into a source object.
    We additionally make calculations for the ranges. And finally add
    the needed glyphs (markers) taking the references from the source.

    """

    glyph = AreaGlyph
